@@ -1,10 +1,15 @@
 import asyncio
-from typing import Awaitable, List
+import logging
+from typing import Any, Awaitable, List, NamedTuple
 
-from .gear import DaliGear
+from .address import AbstractDaliAddress, DaliGearAddress, DaliGearGroupAddress
+from .gear import DaliGear, DaliGearGroup
 from .types import (DaliCommandCode, DaliException, FramingException,
                     MessageSource, SearchAddressClashException,
-                    SpecialCommandCode)
+                    SpecialCommandCode, ItemDelta)
+
+
+_LOGGER = logging.getLogger(__name__)
 
 
 class SearchAddressSender:
@@ -22,34 +27,96 @@ class SearchAddressSender:
         h = (addr >> 16) & 0xFF
 
         if l != self.lastL:
-            await self.drv.send_special_cmd(DaliCommandCode.SearchAddrL, l)
+            await self.drv.send_special_cmd(SpecialCommandCode.SearchAddrL, l)
             self.lastL = l
         if m != self.lastM:
-            await self.drv.send_special_cmd(DaliCommandCode.SearchAddrM, m)
+            await self.drv.send_special_cmd(SpecialCommandCode.SearchAddrM, m)
             self.lastM = m
         if h != self.lastH:
-            await self.drv.send_special_cmd(DaliCommandCode.SearchAddrH, h)
+            await self.drv.send_special_cmd(SpecialCommandCode.SearchAddrH, h)
             self.lastH = h
 
 
-
 class DaliBusTransciever:
+    devices: List[DaliGear]
+    groups: List[DaliGearGroup]
+    transcievers: List['DaliBusTransciever'] = []
+    manufacturer: str|None
+    model: str|None
+    serial: str|None
     
     def __init__(self) -> None:
         self._new_message_callbacks = []
+        self.devices = [DaliGear(self, DaliGearAddress(index)) for index in range(0,64)]
+        self.groups = [DaliGearGroup(self, DaliGearGroupAddress(index)) for index in range(0,16)]
+        self.manufacturer = None
+        self.model = None
+        self.serial = None
+
+
+    @property
+    def human_readable_name(self):
+        if self.manufacturer is not None and self.model is not None and self.serial is not None:
+            return "{} {} ({})".format(self.manufacturer, self.model, self.serial)
+        return str(self)
+
 
     @property
     def unique_id(self):
         raise Exception("Not Implemented")
 
+    @property
+    def present_gear(self):
+        return [g for g in self.devices if g.present]
+
+    @property
+    def present_groups(self):
+        return [g for g in self.groups if g.has_gear]
+
+
+    @classmethod
+    async def for_unique_id(cls, transciever_id: str) -> 'DaliBusTransciever':
+        try:
+            return next(tx for tx in DaliBusTransciever.transcievers if tx.unique_id == transciever_id) 
+        except StopIteration:
+            return None
+
+    @classmethod
+    def find_gear(cls, unique_id):
+        """Finds the gear by its unique Id, no matter which bus it is on.  This allows for devices 
+        to move around, and to change short address.  This does mean that the device_info via-device 
+        might change after a bus re-order"""
+        for tx in DaliBusTransciever.transcievers:
+            for gear in tx.devices:
+                if gear.present and gear.unique_id == unique_id:
+                    return gear
+            for group in tx.groups:
+                if group.present and group.unique_id == unique_id:
+                    return group
+
+
+
+    @classmethod
+    async def scan_for_transcievers(cls) -> ItemDelta['DaliBusTransciever']:
+        subclz = cls.__subclasses__()
+
+        res = ItemDelta([], [])
+        for sc in subclz:
+            delta = await sc._transciever_scan()
+            res.extend(delta)
+            cls.transcievers.extend(delta.added)
+            for to_remove in delta.removed:
+                cls.transcievers.remove(to_remove)
+        return res
+
     def _send(self, data: int, length: int = 16, repeat: int = 1)-> Awaitable[int|None]:
         raise Exception("Not Implemented")
 
-    def send_direct_arc_power(self, address: int, level) -> Awaitable[None]:
-        return self._send((address << 8) | level)
+    def send_direct_arc_power(self, address: AbstractDaliAddress, level) -> Awaitable[None]:
+        return self._send((address.code << 8) | level)
 
-    def send_cmd(self, address: int, cmd: DaliCommandCode, repeat=1) -> Awaitable[int|None]:
-        return self._send((address << 9 ) | 0x0100 | cmd.value, repeat=repeat)
+    def send_cmd(self, address: AbstractDaliAddress, cmd: DaliCommandCode, repeat=1) -> Awaitable[int|None]:
+        return self._send((address.code << 8) | 0x0100 | cmd.value, repeat=repeat)
 
     def send_special_cmd(self, special_cmd: SpecialCommandCode, param: int = 0, repeat=1) -> Awaitable[None]:
         return self._send((special_cmd.value << 8) | param, repeat=repeat)
@@ -78,15 +145,16 @@ class DaliBusTransciever:
         pass # By default, do nothing
 
     async def scan_for_gear(self):
-        self.devices = []
-        for address in range(0,64):
-            gear = DaliGear(self, address)
-            await gear.fetch_deviceinfo()
+        for gear in self.devices:
+            _LOGGER.debug("Scanning %s", gear)
+            await gear.update_all()
+
+        for group in self.groups:
+            await group.update_all()
 
-            if gear.device_type:
-                self.devices.append(gear)
         return self.devices
 
+
     async def compare(self, search, address_sender):
         """
         Compares the supplied search value with items on the bus. 
@@ -97,7 +165,7 @@ class DaliBusTransciever:
         # print("Compare 0x{:06x}".format(search))
         await address_sender.send(search)
         try:
-            found = await self.send_special_cmd(DaliCommandCode.Compare)
+            found = await self.send_special_cmd(SpecialCommandCode.Compare)
             if found == 0xFF:
                 # Precisely one device is equal to or less than high
                 return 1
@@ -141,35 +209,32 @@ class DaliBusTransciever:
                 # 1 or more devices below or equal to mid, but we don't know which
                 high = mid
 
-        
-
-
 
     async def commission(self):
         # Terminate any outstanding initialise.
-        await self.send_special_cmd(DaliCommandCode.Terminate, 0)
+        await self.send_special_cmd(SpecialCommandCode.Terminate, 0)
         try:
 
             # TODO start quiescent mode (24 bit command.)
 
             # Put devices in initialisation mode. 
-            await self.send_special_cmd(DaliCommandCode.Initialise, repeat=2)
+            await self.send_special_cmd(SpecialCommandCode.Initialise, repeat=2)
 
 
             # Clear out any existing short addresses
-            await self.send_special_cmd(DaliCommandCode.SetDTR0, 0xFF)
+            await self.send_special_cmd(SpecialCommandCode.SetDTR0, 0xFF)
             await self.broadcast(DaliCommandCode.SetShortAddress, repeat=2)
 
             # Reset operating mode
-            await self.send_special_cmd(DaliCommandCode.SetDTR0, 128)
+            await self.send_special_cmd(SpecialCommandCode.SetDTR0, 128)
             await self.broadcast(DaliCommandCode.SetOperatingMode, repeat=2)
 
             # Remove devices from groups
             for group in range(16):
-                await self.broadcast(DaliCommandCode.RemoveFromGroup | group, repeat=2)
+                await self.broadcast(DaliCommandCode(DaliCommandCode.RemoveFromGroup0.value + group), repeat=2)
 
             # Randomise the search addresses for all devices. 
-            await self.send_special_cmd(DaliCommandCode.Randomise, repeat=2)
+            await self.send_special_cmd(SpecialCommandCode.Randomise, repeat=2)
             await asyncio.sleep(0.1)  
 
             
@@ -187,12 +252,12 @@ class DaliBusTransciever:
                     if found:
                         short_addr = available_short_addresses.pop(0)
                         shifted = (short_addr << 1) | 0x01
-                        await self.send_special_cmd(DaliCommandCode.ProgramShortAddress, shifted)
-                        queried_short_addr = await self.send_special_cmd(DaliCommandCode.QueryShortAddress)
+                        await self.send_special_cmd(SpecialCommandCode.ProgramShortAddress, shifted)
+                        queried_short_addr = await self.send_special_cmd(SpecialCommandCode.QueryShortAddress)
 
                         if queried_short_addr == shifted:
                             # Good, the device took the address
-                            await self.send_special_cmd(DaliCommandCode.Withdraw)
+                            await self.send_special_cmd(SpecialCommandCode.Withdraw)
                         else:
                             raise DaliException("Short Address did not stick (Returned {:02x} instead of {:02x})".format(queried_short_addr, shifted))
                         search_floor = found + 1
@@ -204,15 +269,11 @@ class DaliBusTransciever:
                     search_floor = 0
         finally:
             # Make sure we've terminated our commission process
-            await self.send_special_cmd(DaliCommandCode.Terminate, 0)
+            await self.send_special_cmd(SpecialCommandCode.Terminate, 0)
                 
-
-def is_address_group(addr: int):
-    return addr & 0x80 != 0
-
-
-def filter_for_address(entities, address):
-    return [e for e in entities if e.matches_address(address)]
+class GearListAndAttribute(NamedTuple):
+    gear: List[DaliGear]
+    attribute: str
 
 
 class DaliMessage: 
@@ -228,48 +289,43 @@ class DaliMessage:
 
     def __repr__(self):
         return "{} ({})".format(self.source.name, self.sequence_number)
+
+    @property
+    def is_response(self):
+        return False
+
     
     @property
-    def affected_gear(self):
+    def affected_gear(self) -> GearListAndAttribute | None:
         """Each message may affect one or more gear on its bus.  This property tells us
         which (already discovered) gear would be affected"""
-        return []
+        return None
+
+    def get_affected_groups(self, affected_gear) -> List[DaliGearGroup]:
+        """Once we've worked out which devices are affected by a message, we need to work out
+        which groups also incidentally get affected."""
+
+        mask = 0
+        for g in affected_gear:
+            mask = mask | g.groups
+
+        return [g for g in self.transciever.groups if mask & g.mask != 0]
+
+
 
 class ResponseMessage(DaliMessage):
-    pass
+    @property
+    def is_response(self):
+        return True
 
 class NakMessage(ResponseMessage):
     def __repr__(self):
         return super().__repr__() + " NAK"
 
-class AddressedMessage(DaliMessage):
-    address: int
-
-    def __init__(self, transciever: DaliBusTransciever, source: MessageSource, sequence_number: int, address: int):
-        super().__init__(transciever, source, sequence_number)
-        self.address = address
-
-    @property
-    def affected_gear(self):
-        return [g for g in self.transciever.devices if g.matches_address(self.address)]
-
-
-
-class DirectArcPowerCommand(AddressedMessage):
-    value: int
-
-    def __init__(self, transciever: DaliBusTransciever, source: MessageSource, sequence_number: int, address: int, value: int):
-        super().__init__(transciever, source, sequence_number, address)
-        self.value = value
 
+class BadFrame(ResponseMessage):
     def __repr__(self):
-        if is_address_group(self.address):
-            return super().__repr__() + " DAPC(G{}, {})".format(self.address >> 1, self.value)
-        else:
-            return super().__repr__() + " DAPC(A{}, {})".format(self.address >> 1, self.value)
-
-
-
+        return super.__repr__() + " Bad Frame"
 
 class NumericResponseMessage(ResponseMessage):
     value: int
@@ -280,25 +336,54 @@ class NumericResponseMessage(ResponseMessage):
     def __repr__(self):
         return super().__repr__() + " < {}".format(self.value)
 
+
+
+
+class AddressedMessage(DaliMessage):
+    address: AbstractDaliAddress
+
+    def __init__(self, transciever: DaliBusTransciever, source: MessageSource, sequence_number: int, address: AbstractDaliAddress):
+        super().__init__(transciever, source, sequence_number)
+        self.address = address
+
+    @property
+    def affected_gear(self) -> GearListAndAttribute:
+        return GearListAndAttribute([g for g in self.transciever.devices if self.address.matches_gear(g)], "level")
+
+
+
+
+
+class DirectArcPowerCommand(AddressedMessage):
+    value: int
+
+    def __init__(self, transciever: DaliBusTransciever, source: MessageSource, sequence_number: int, address: AbstractDaliAddress, value: int):
+        super().__init__(transciever, source, sequence_number, address)
+        self.value = value
+
+    def __repr__(self):
+        return super().__repr__() + " DAPC({}, {})".format(self.address, self.value)
+
+
+
+
 class AddressedCommand(AddressedMessage):
     command: DaliCommandCode    
 
-    def __init__(self, transciever: DaliBusTransciever, source: MessageSource, sequence_number: int, address: int, command: DaliCommandCode):
+    def __init__(self, transciever: DaliBusTransciever, source: MessageSource, sequence_number: int, address: AbstractDaliAddress, command: DaliCommandCode):
         super().__init__(transciever, source, sequence_number, address)
         self.command = command
 
     def __repr__(self):
-        if is_address_group(0x80):
-            return super().__repr__() + " > {}(A{})".format(self.command.name, self.address >> 1)
-        else:
-            return super().__repr__() + " > {}(G{})".format(self.command.name, self.address >> 1)
+        return super().__repr__() + " > {}({})".format(self.command.name, self.address)
 
     @property
-    def affected_gear(self):
-        if self.command.has_side_effects:
-            return super().affected_gear
+    def affected_gear(self) -> GearListAndAttribute:
+        side_effect = self.command.side_effect
+        if side_effect is not None:
+            return GearListAndAttribute(super().affected_gear.gear, side_effect)
         else:
-            return []
+            return None
 
 
 class SpecialCommand(DaliMessage):
@@ -314,26 +399,3 @@ class SpecialCommand(DaliMessage):
         return super().__repr__() + " special > {}(0x{:02x})".format(self.command.name, self.operand)
 
 
-class BroadcastCommand(DaliMessage):
-    unaddressed: bool = False
-    command: DaliCommandCode
-
-    def __init__(self, transciever: DaliBusTransciever, source: MessageSource, sequence_number: int, unaddressed: bool, command: DaliCommandCode):
-        super().__init__(transciever, source, sequence_number)
-        self.command = command
-        self.unaddressed = unaddressed
-
-    def __repr__(self):
-        if self.unaddressed:
-            return super().__repr__() + " unaddressed broadcast {}".format(self.command.name)
-        else:
-            return super().__repr__() + " broadcast {}".format(self.command.name)
-
-    @property
-    def affected_gear(self):
-        if self.command.has_side_effects:
-            return self.transciever.devices
-
-class BadFrame(DaliMessage):
-    def __repr__(self):
-        return super.__repr__() + " Bad Frame"
